@@ -9,10 +9,62 @@ from decimal import Decimal as Decimal
 from datetime import date, datetime, timedelta, time as dt_time
 from json import dumps
 from uuid import UUID
+from django.contrib.postgres.fields import (HStoreField, ArrayField, IntegerRangeField,
+    BigIntegerRangeField, DecimalRangeField, DateTimeRangeField, DateRangeField)
+from psycopg2.extras import Range
 
 
-# TODO: postgres custom field support: arrays (cumbersome), hstore, range fields (easy)
-# TODO: transport encoding not tested yet (can we derived default from psycopg2?)
+# TODO: typings, docs cleanup
+# TODO: import guard in query.py
+
+
+# postgres connection encodings mapped to python
+# taken from https://www.postgresql.org/docs/current/multibyte.html
+# python counterparts shamelessly taken from psycopg3 :)
+CONNECTION_ENCODINGS = {
+    'BIG5': 'big5',
+    'EUC_CN': 'gb2312',
+    'EUC_JP': 'euc_jp',
+    'EUC_JIS_2004': 'euc_jis_2004',
+    'EUC_KR': 'euc_kr',
+    #'EUC_TW': not supported
+    'GB18030': 'gb18030',
+    'GBK': 'gbk',
+    'ISO_8859_5': 'iso8859-5',
+    'ISO_8859_6': 'iso8859-6',
+    'ISO_8859_7': 'iso8859-7',
+    'ISO_8859_8': 'iso8859-8',
+    'JOHAB': 'johab',
+    'KOI8R': 'koi8-r',
+    'KOI8U': 'koi8-u',
+    'LATIN1': 'iso8859-1',
+    'LATIN2': 'iso8859-2',
+    'LATIN3': 'iso8859-3',
+    'LATIN4': 'iso8859-4',
+    'LATIN5': 'iso8859-9',
+    'LATIN6': 'iso8859-10',
+    'LATIN7': 'iso8859-13',
+    'LATIN8': 'iso8859-14',
+    'LATIN9': 'iso8859-15',
+    'LATIN10': 'iso8859-16',
+    #'MULE_INTERNAL': not supported
+    'SJIS': 'shift_jis',
+    'SHIFT_JIS_2004': 'shift_jis_2004',
+    'SQL_ASCII': 'ascii',
+    'UHC': 'cp949',
+    'UTF8': 'utf-8',
+    'WIN866': 'cp866',
+    'WIN874': 'cp874',
+    'WIN1250': 'cp1250',
+    'WIN1251': 'cp1251',
+    'WIN1252': 'cp1252',
+    'WIN1253': 'cp1253',
+    'WIN1254': 'cp1254',
+    'WIN1255': 'cp1255',
+    'WIN1256': 'cp1256',
+    'WIN1257': 'cp1257',
+    'WIN1258': 'cp1258',
+}
 
 
 # NULL placeholder for COPY FROM
@@ -22,7 +74,7 @@ LAZY_PLACEHOLDER = '\x00'
 LAZY_PLACEHOLDER_BYTE = b'\x00'
 
 
-def textEscape(v):
+def text_escape(v):
     """
     Escape str-like data for postgres' TEXT format.
     """
@@ -35,7 +87,8 @@ def Int(v, lazy):
     """Test and pass along ``int``, raise for any other."""
     if isinstance(v, int):
         return v
-    raise TypeError('expected int type')
+    raise TypeError(f'expected type {int}')
+Int.array_escape = False
 
 
 def IntOrNone(v, lazy):
@@ -44,7 +97,8 @@ def IntOrNone(v, lazy):
         return NULL
     if isinstance(v, int):
         return v
-    raise TypeError('expected int or NoneType')
+    raise TypeError(f'expected type {int} or None')
+IntOrNone.array_escape = False
 
 
 def _lazy_binary(f, v):
@@ -73,7 +127,8 @@ def Binary(v, lazy):
             lazy.append((_lazy_binary, v))
             return '\\\\x' + LAZY_PLACEHOLDER
         return '\\\\x' + v.hex()
-    raise TypeError('expected memoryview or bytes type')
+    raise TypeError(f'expected types {memoryview} or {bytes}')
+Binary.array_escape = True
 
 
 def BinaryOrNone(v, lazy):
@@ -85,14 +140,16 @@ def BinaryOrNone(v, lazy):
             lazy.append((_lazy_binary, v))
             return '\\\\x' + LAZY_PLACEHOLDER
         return '\\\\x' + v.hex()
-    raise TypeError('expected memoryview, bytes or NoneType')
+    raise TypeError(f'expected types {memoryview}, {bytes} or None')
+BinaryOrNone.array_escape = True
 
 
 def Boolean(v, lazy):
     """Test and pass along ``bool``, raise for any other."""
     if isinstance(v, bool):
         return v
-    raise TypeError('expected bool type')
+    raise TypeError(f'expected type {bool}')
+Boolean.array_escape = False
 
 
 def BooleanOrNone(v, lazy):
@@ -101,14 +158,16 @@ def BooleanOrNone(v, lazy):
         return NULL
     if isinstance(v, bool):
         return v
-    raise TypeError('expected bool or NoneType')
+    raise TypeError(f'expected type {bool} or None')
+BooleanOrNone.array_escape = False
 
 
 def Date(v, lazy):
     """Test and pass along ``datetime.date``, raise for any other."""
     if isinstance(v, date):
         return v
-    raise TypeError('expected datetime.date type')
+    raise TypeError(f'expected type {date}')
+Date.array_escape = False
 
 
 def DateOrNone(v, lazy):
@@ -117,14 +176,16 @@ def DateOrNone(v, lazy):
         return NULL
     if isinstance(v, date):
         return v
-    raise TypeError('expected datetime.date or NoneType')
+    raise TypeError(f'expected type {date} or None')
+DateOrNone.array_escape = False
 
 
 def Datetime(v, lazy):
     """Test and pass along ``datetime``, raise for any other."""
     if isinstance(v, datetime):
         return v
-    raise TypeError('expected datetime type')
+    raise TypeError(f'expected type {datetime}')
+Datetime.array_escape = True
 
 
 def DatetimeOrNone(v, lazy):
@@ -133,14 +194,16 @@ def DatetimeOrNone(v, lazy):
         return NULL
     if isinstance(v, datetime):
         return v
-    raise TypeError('expected datetime or NoneType')
+    raise TypeError(f'expected type {datetime} or None')
+DatetimeOrNone.array_escape = True
 
 
 def Numeric(v, lazy):
     """Test and pass along ``Decimal``, raise for any other."""
     if isinstance(v, Decimal):
         return v
-    raise TypeError('expected Decimal type')
+    raise TypeError(f'expected type {Decimal}')
+Numeric.array_escape = False
 
 
 def NumericOrNone(v, lazy):
@@ -149,14 +212,16 @@ def NumericOrNone(v, lazy):
         return NULL
     if isinstance(v, Decimal):
         return v
-    raise TypeError('expected Decimal or NoneType')
+    raise TypeError(f'expected type {Decimal} or None')
+NumericOrNone.array_escape = False
 
 
 def Duration(v, lazy):
     """Test and pass along ``timedelta``, raise for any other."""
     if isinstance(v, timedelta):
         return v
-    raise TypeError('expected timedelta type')
+    raise TypeError(f'expected type {timedelta}')
+Duration.array_escape = True
 
 
 def DurationOrNone(v, lazy):
@@ -165,14 +230,16 @@ def DurationOrNone(v, lazy):
         return NULL
     if isinstance(v, timedelta):
         return v
-    raise TypeError('expected timedelta or NoneType')
+    raise TypeError(f'expected type {timedelta} or None')
+DurationOrNone.array_escape = True
 
 
 def Float(v, lazy):
     """Test and pass along ``float`` or ``int``, raise for any other."""
     if isinstance(v, (float, int)):
         return v
-    raise TypeError('expected float or int type')
+    raise TypeError(f'expected types {float} or {int}')
+Float.array_escape = False
 
 
 def FloatOrNone(v, lazy):
@@ -181,16 +248,19 @@ def FloatOrNone(v, lazy):
         return NULL
     if isinstance(v, (float, int)):
         return v
-    raise TypeError('expected float, int or NoneType')
+    raise TypeError(f'expected types {float}, {int} or None')
+FloatOrNone.array_escape = False
 
 
+# TODO: test and document Json vs. JsonOrNone behavior (sql null vs. json null)
 def Json(v, lazy):
     """
     Default JSON encoder using ``json.dumps``.
 
     This version encodes ``None`` as json null value.
     """
-    return textEscape(dumps(v))
+    return text_escape(dumps(v))
+Json.array_escape = True
 
 
 def JsonOrNone(v, lazy):
@@ -201,7 +271,8 @@ def JsonOrNone(v, lazy):
     """
     if v is None:
         return NULL
-    return textEscape(dumps(v))
+    return text_escape(dumps(v))
+JsonOrNone.array_escape = True
 
 
 def Text(v, lazy):
@@ -212,8 +283,9 @@ def Text(v, lazy):
     for the TEXT format of COPY FROM.
     """
     if isinstance(v, str):
-        return textEscape(v)
-    raise TypeError('expected str type')
+        return text_escape(v)
+    raise TypeError(f'expected type {str}')
+Text.array_escape = True
 
 
 def TextOrNone(v, lazy):
@@ -221,15 +293,17 @@ def TextOrNone(v, lazy):
     if v is None:
         return NULL
     if isinstance(v, str):
-        return textEscape(v)
-    raise TypeError('expected str or NoneType')
+        return text_escape(v)
+    raise TypeError(f'expected type {str} or None')
+TextOrNone.array_escape = True
 
 
 def Time(v, lazy):
     """Test and pass along ``datetime.time``, raise for any other."""
     if isinstance(v, dt_time):
         return v
-    raise TypeError('expected datetime.time type')
+    raise TypeError(f'expected type {dt_time}')
+Time.array_escape = False
 
 
 def TimeOrNone(v, lazy):
@@ -238,14 +312,16 @@ def TimeOrNone(v, lazy):
         return NULL
     if isinstance(v, dt_time):
         return v
-    raise TypeError('expected datetime.time or NoneType')
+    raise TypeError(f'expected type {dt_time} or None')
+TimeOrNone.array_escape = False
 
 
 def Uuid(v, lazy):
     """Test and pass along ``UUID``, raise for any other."""
     if isinstance(v, UUID):
         return v
-    raise TypeError('expected UUID type')
+    raise TypeError(f'expected type {UUID}')
+Uuid.array_escape = False
 
 
 def UuidOrNone(v, lazy):
@@ -254,7 +330,109 @@ def UuidOrNone(v, lazy):
         return NULL
     if isinstance(v, UUID):
         return v
-    raise TypeError('expected UUID or NoneType')
+    raise TypeError(f'expected type {UUID} or None')
+UuidOrNone.array_escape = False
+
+
+"""
+Special handling of nested types
+
+Nested types behave way different in COPY FROM TEXT format,
+than on top level (kinda falling back on SQL syntax format).
+From the tests this applies to values in arrays and hstore
+(prolly applies to all custom composite types, not tested).
+
+Rules for nested types:
+- a backslash in nested strings needs 4x escaping, e.g. \ --> \\\\
+- nested str values may need explicit quoting --> always quoted
+- due to quoting, " needs \\" escape in nested strings
+- null value is again sql NULL
+"""
+def quote(v):
+    return '"' + v.replace('"', '\\\\"') + '"'
+
+def text_escape_nested(v):
+    """
+    Escape nested str-like data for postgres' TEXT format.
+    The nested variant is needed for array and hstore data
+    (prolly for any custom composite types, untested).
+    """
+    return (v.replace('\\', '\\\\\\\\')
+        .replace('\b', '\\b').replace('\f', '\\f').replace('\n', '\\n')
+        .replace('\r', '\\r').replace('\t', '\\t').replace('\v', '\\v'))
+
+SQL_NULL = 'NULL'
+
+def array_escape(v):
+    return '"' + v.replace('\\\\', '\\\\\\\\').replace('"', '\\\\"') + '"'
+
+
+def HStore(v, lazy):
+    """
+    HStore field encoder. Expects a ``dict`` as input type
+    with str keys and str|None values. Any other types will raise.
+
+    The generated TEXT format representation is always quoted
+    in the form: ``"key"=>"value with \\"double quotes\\""``.
+    """
+    if isinstance(v, dict):
+        parts = []
+        for k, v in v.items():
+            if not isinstance(k, str):
+                raise TypeError(f'expected type {str} for keys')
+            if v is not None and not isinstance(v, str):
+                raise TypeError(f'expected type {str} or None for values')
+            parts.append(
+                f'{quote(text_escape_nested(k))}=>'
+                f'{SQL_NULL if v is None else quote(text_escape_nested(v))}'
+            )
+        return ','.join(parts)
+    raise TypeError(f'expected type {dict}')
+HStore.array_escape = True
+
+
+def HStoreOrNone(v, lazy):
+    """Same as ``Hstore``, additionally handling ``None`` as NULL."""
+    if v is None:
+        return NULL
+    if isinstance(v, dict):
+        return HStore(v, lazy)
+    raise TypeError(f'expected type {dict} or None')
+HStoreOrNone.array_escape = True
+
+
+def range_factory(_type):
+    def wrap(v, lazy):
+        if isinstance(v, Range) and all(map(lambda e: isinstance(e, _type), (v.lower, v.upper))):
+            return v
+        raise TypeError(f'expected type {_type}')
+    wrap.array_escape = True
+    def wrap_none(v, lazy):
+        if v is None:
+            return NULL
+        if isinstance(v, Range) and all(map(lambda e: isinstance(e, _type), (v.lower, v.upper))):
+            return v
+        raise TypeError(f'expected type {_type} or None')
+    wrap_none.array_escape = True
+    return wrap, wrap_none
+
+
+# TODO: cache encoder for later calls?
+# NOTE: array descent only happens on lists (to allow tuples as json native)
+# FIXME: should we respect null field setting in dim descent?
+def array_factory(field):
+    base = field.base_field
+    while isinstance(base, ArrayField):
+        base = base.base_field
+    enc = get_encoder(base)
+    def wrap(v, lazy, dim=0):
+        if v is None:
+            return SQL_NULL if dim else NULL
+        if isinstance(v, (list,)):
+            return f'{{{",".join(wrap(e, lazy, dim+1) for e in v)}}}'
+        final = str(enc(v, lazy))
+        return array_escape(final) if enc.array_escape else final
+    return wrap
 
 
 ENCODERS = {
@@ -286,7 +464,14 @@ ENCODERS = {
     models.TimeField: (Time, TimeOrNone),
     models.URLField: (Text, TextOrNone),
     models.UUIDField: (Uuid, UuidOrNone),
-    #ArrayField, HStore, Range ...
+    # postgres specific fields
+    HStoreField: (HStore, HStoreOrNone),
+    IntegerRangeField: range_factory(int),
+    BigIntegerRangeField: range_factory(int),
+    DecimalRangeField: range_factory(Decimal),
+    DateTimeRangeField: range_factory(datetime),
+    DateRangeField: range_factory(date),
+    # ArrayField: handled by array_factory(field) in get_encoder
 }
 
 
@@ -307,6 +492,8 @@ def get_encoder(field):
     """Get registered encoder for field."""
     if field.is_relation:
         return get_encoder(field.target_field)
+    if isinstance(field, ArrayField):
+        return array_factory(field)
     for cls in type(field).__mro__:
         enc = ENCODERS.get(cls)
         if enc:
@@ -385,7 +572,7 @@ def copy_from(c, tname, data, columns, get, encs, encoding):
         f.close()
 
 
-def prepare_create_columns(column_def):
+def create_columns(column_def):
     """
     Prepare columns for table create as follows:
     - types copied from target table
@@ -408,8 +595,8 @@ def copy_update(
     qs: models.QuerySet,
     objs: Sequence[models.Model],
     fieldnames: Sequence[str],
-    transport_encoding: Optional[str] = 'utf-8',
-    field_encoders: Optional[Dict[str, Any]] = None
+    field_encoders: Optional[Dict[str, Any]] = None,
+    encoding: Optional[str] = None
 ) -> int:
     qs._for_write = True
     conn = connections[qs.db]
@@ -442,8 +629,9 @@ def copy_update(
     with transaction.atomic(using=conn.alias, savepoint=False), conn.cursor() as c:
         temp = f'temp_cu_{model._meta.db_table}'
         c.execute(f'DROP TABLE IF EXISTS "{temp}"')
-        c.execute(f'CREATE TEMPORARY TABLE "{temp}" ({prepare_create_columns(column_def)})')
-        copy_from(c, temp, objs, colnames, get, encs, transport_encoding)
+        c.execute(f'CREATE TEMPORARY TABLE "{temp}" ({create_columns(column_def)})')
+        copy_from(c, temp, objs, colnames, get, encs,
+            encoding or CONNECTION_ENCODINGS[c.connection.encoding])
         c.execute(f'ANALYZE "{temp}" ({pk_field.column})')
         c.execute(update_sql(model._meta.db_table, temp, pk_field.column, fields))
         rows_updated = c.rowcount
